from typing import Dict, Any, MutableMapping, ClassVar, Type

from venom.common import IntegerValueConverter, BooleanValueConverter, DateTimeConverter, DateConverter
from venom.common import StringValueConverter, NumberValueConverter
from venom.rpc.context import RequestContextDescriptor
from venom.rpc.method import Method, MethodDescriptor
from venom.util import meta, MetaDict


class ServiceManager(object):
    def __init__(self, meta: MetaDict, meta_changes: MetaDict):
        self.meta = meta

    @staticmethod
    def get_service_name(cls_name: str) -> str:
        name = cls_name.lower()
        for postfix in ('service', 'stub'):
            if name.endswith(postfix):
                return name[:-len(postfix)]
        return name

    @classmethod
    def prepare_meta(cls, name: str, meta: MetaDict, meta_changes: MetaDict) -> MetaDict:
        if not meta_changes.get('name'):
            meta.name = cls.get_service_name(name)

        if not meta_changes.get('http_path', None):
            meta.http_path = f"/{meta.name.lower().replace('_', '-')}"
        return meta

    def prepare_method(self, service: Type['Service'], method: MethodDescriptor, name: str):
        return method.prepare(service, name)

    def register(self, venom: 'venom.Venom') -> None:
        pass


class ServiceMeta(type):
    def __new__(metacls, what, bases=(), members=None):
        meta_, meta_changes = meta(bases, members)
        meta_ = meta_.manager.prepare_meta(what, meta_, meta_changes)
        manager = meta_.manager(meta_, meta_changes)
        method_descriptors = {}

        for base in bases:
            if isinstance(base, ServiceMeta):
                method_descriptors.update(base.__method_descriptors__)

        for name, member in members.items():
            if isinstance(member, MethodDescriptor):
                method_descriptors[name] = member

        stub = meta_changes.get('stub')
        if stub:
            if not isinstance(stub, type) and issubclass(stub, Service):
                raise TypeError('Meta.stub must be a Service')

        members['__manager__'] = manager
        members['__meta__'] = manager.meta
        members['__method_descriptors__'] = method_descriptors
        members['__stub__'] = stub
        cls = super(ServiceMeta, metacls).__new__(metacls, name, bases, members)

        if stub:
            for name, method in stub.__method_descriptors__.items():
                if name not in method_descriptors:
                    method_descriptors[name] = method

        cls.__methods__ = methods = {
            name: manager.prepare_method(cls, method, name)
            for name, method in method_descriptors.items()
        }

        for name, method in methods.items():
            setattr(cls, name, method)

        return cls


class Service(object, metaclass=ServiceMeta):
    """
    A service is a collection of functions, possibly HTTP routes.
    """
    __meta__: ClassVar['venom.util.AttributeDict'] = None
    __manager__: ClassVar[ServiceManager] = None
    __method_descriptors__: ClassVar[Dict[str, MethodDescriptor]] = None
    __methods__: ClassVar[Dict[str, Method]] = None
    __stub__: ClassVar[Type['Service']] = None

    context = RequestContextDescriptor()

    def __init__(self, venom: 'venom.rpc.Venom' = None) -> None:
        self.venom = venom

    class Meta:
        name = None
        manager = ServiceManager
        messages = ()
        converters = (
            StringValueConverter,
            IntegerValueConverter,
            NumberValueConverter,
            BooleanValueConverter,
            DateTimeConverter,
            DateConverter)
        stub = None
        http_path = None

    def __repr__(self):
        return f'<Service [{self.__meta__.name}]>'



from functools import partial

import asyncio
from threading import Thread
from typing import Type

from venom.rpc.comms import AbstractClient
from venom.protocol import Protocol, JSON

try:
    from grpc.beta import implementations
    from grpc.framework.interfaces.face import utilities
except ImportError:
    raise RuntimeError("You must install the 'grpcio' package to use the GRPC features of Venom RPC")


# TODO WARNING: GRPC Core is not asynchronous (as of October 2016). Therefore all calls made to the server are blocking.

def create_server(venom: 'venom.rpc.Venom',
                  *,
                  protocol_factory: Type[Protocol] = JSON,
                  pool=None,
                  pool_size=None,
                  default_timeout=None,
                  maximum_timeout=None,
                  loop=None):
    if loop is None:
        loop = asyncio.new_event_loop()

    request_deserializers = {}
    response_serializers = {}
    method_implementations = {}

    def grpc_unary_unary(rpc, venom, service, request, context, *, loop):
        instance = venom.get_instance(service)
        future = asyncio.run_coroutine_threadsafe(rpc.invoke(instance, request), loop)
        # TODO use context.timeout
        return future.result()

    for service, rpc in venom.iter_methods():
        grpc_name = (service.__meta__.name, rpc.name)
        request_deserializers[grpc_name] = protocol_factory(rpc.request).unpack
        response_serializers[grpc_name] = protocol_factory(rpc.response).pack
        method_implementations[grpc_name] = utilities.unary_unary_inline(partial(grpc_unary_unary,
                                                                                 rpc,
                                                                                 venom,
                                                                                 service,
                                                                                 loop=loop))

    server_options = implementations.server_options(request_deserializers=request_deserializers,
                                                    response_serializers=response_serializers,
                                                    thread_pool=pool, thread_pool_size=pool_size,
                                                    default_timeout=default_timeout,
                                                    maximum_timeout=maximum_timeout)

    def event_loop_runner(loop):
        asyncio.set_event_loop(loop)
        loop.run_forever()

    t = Thread(target=event_loop_runner, args=(loop,))
    t.daemon = True
    t.start()

    return implementations.server(method_implementations, options=server_options)


class Client(AbstractClient):
    def __init__(self, stub: Type['venom.rpc.Service'], host=None, port=50051, *, protocol_factory: Type[Protocol] = None):
        super().__init__(stub, protocol_factory=protocol_factory)
        channel = implementations.insecure_channel(host, port)
        self._group = stub.__meta__.name
        self._grpc_stub = self._create_grpc_stub(stub, channel)

    def _create_grpc_stub(self, stub, channel, host=None, metadata_transformer=None, pool=None, pool_size=None):
        request_serializers = {}
        response_deserializers = {}

        for rpc in stub.__methods__.values():
            request_serializers[(self._group, rpc.name)] = self._protocol_factory(rpc.request).pack
            response_deserializers[(self._group, rpc.name)] = self._protocol_factory(rpc.response).unpack

        stub_options = implementations.stub_options(host=host,
                                                    metadata_transformer=metadata_transformer,
                                                    request_serializers=request_serializers,
                                                    response_deserializers=response_deserializers,
                                                    thread_pool=pool,
                                                    thread_pool_size=pool_size)

        return implementations.generic_stub(channel, stub_options)

    async def invoke(self,
                     rpc: 'venom.stub.RPC',
                     request: 'venom.message.Message',
                     *,
                     context: 'venom.rpc.RequestContext' = None,
                     loop: asyncio.AbstractEventLoop = None,
                     timeout: int = None):
        if loop is None:
            loop = asyncio.get_event_loop()

        future = loop.run_in_executor(None, partial(self._grpc_stub.blocking_unary_unary,
                                                    self._group,
                                                    rpc.name,
                                                    request,
                                                    timeout=timeout))

        return await future



from abc import ABCMeta
from importlib import import_module
from typing import TypeVar, Generic, Any, Union, Type

import collections

from venom.util import cached_property, AttributeDict, camelcase

T = TypeVar('T', bool, int, float, str, bytes, 'venom.message.Message')


class FieldDescriptor(Generic[T], metaclass=ABCMeta):
    _name: str = None
    _json_name: str = None

    def __init__(self, name: str = None, *, json_name: str = None):
        self._name = name
        self._json_name = json_name

    def __get__(self, instance: 'venom.message.Message', owner):
        if instance is None:
            return self
        try:
            return instance[self.name]
        except KeyError:
            return self.default()

    def default(self):
        return None

    def __set_name__(self, owner, name):
        if self._name is None:
            self._name = name

    @property
    def json_name(self):
        if self._json_name is None:
            return camelcase(self.name)
        return self._json_name

    @property
    def name(self):
        return self._name

    # TODO wait on https://github.com/python/mypy/issues/244
    def __set__(self, instance: 'venom.message.Message', value: T):
        instance[self.name] = value

    # TODO Use Python 3.6 __set_name__()


class Field(Generic[T], FieldDescriptor):
    def __init__(self,
                 type_: Union[Type[T], str],
                 default: Any = None,
                 name: str = None,
                 *,
                 json_name: str = None,
                 **options) -> None:
        super(Field, self).__init__(name, json_name=json_name)
        self._type = type_
        self._default = default
        self.options = AttributeDict(options)

    def default(self):
        if self._default is None:
            return self.type()
        return self._default

    @cached_property
    def type(self) -> Type[T]:
        if isinstance(self._type, str):
            if '.' in self._type:
                module_name, class_name = self._type.rsplit('.', 1)
                module = import_module(module_name)
                return getattr(module, class_name)

            raise RuntimeError('Unable to resolve: {} in {}'.format(self._type, repr(self)))
        return self._type

    def __eq__(self, other):
        if not isinstance(other, Field):
            return False
        return self.type == other.type and self.options == other.options

    def __repr__(self):
        return '<{} {}:{}>'.format(self.__class__.__name__,
                                   self.name,
                                   self._type.__name__ if not isinstance(self._type, str) else repr(self._type))

    def __hash__(self):
        return hash(repr(self))

P = TypeVar('P')


class ConverterField(Generic[T, P], Field[T]):
    def __init__(self,
                 converter: 'venom.converter.Converter[T, P]' = None,
                 **kwargs) -> None:
        super().__init__(converter.wire, **kwargs)
        self.converter = converter

    def __set__(self, instance: T, value: P) -> None:
        instance[self.name] = self.converter.format(value)

    def __get__(self, instance: T, _=None) -> P:
        return self.converter.convert(instance.get(self.name))


class String(Field[str]):
    def __init__(self, **kwargs) -> None:
        super().__init__(str, **kwargs)


class Bytes(Field[bytes]):
    def __init__(self, **kwargs) -> None:
        super().__init__(bytes, **kwargs)


class Bool(Field[bool]):
    def __init__(self, **kwargs) -> None:
        super().__init__(bool, **kwargs)


class Int32(Field[int]):
    def __init__(self, **kwargs) -> None:
        super().__init__(int, **kwargs)


class Int64(Field[int]):
    def __init__(self, **kwargs) -> None:
        super().__init__(int, **kwargs)


Integer = Int = Int64


class Float32(Field[float]):
    def __init__(self, **kwargs) -> None:
        super().__init__(float, **kwargs)


class Float64(Field[float]):
    def __init__(self, **kwargs) -> None:
        super().__init__(float, **kwargs)


Number = Float64

CT = TypeVar('CT', Field, 'MapField', 'RepeatField')


class _RepeatValueProxy(collections.MutableSequence):
    def __init__(self, message: 'venom.message.Message', name: str):
        self.message = message
        self.name = name

    @property
    def _sequence(self) -> list:
        try:
            return self.message[self.name]
        except KeyError:
            return list()

    def __len__(self):
        return len(self._sequence)

    def __getitem__(self, index):
        return self._sequence[index]

    def insert(self, index, value):
        self.message[self.name] = sequence = self._sequence
        sequence.insert(index, value)

    def __delitem__(self, index):
        self.message[self.name] = sequence = self._sequence
        del sequence[index]

    def __setitem__(self, index, value):
        self.message[self.name] = sequence = self._sequence
        sequence[index] = value

    def __iter__(self):
        return iter(self._sequence)


class RepeatField(Generic[CT], FieldDescriptor):
    def __init__(self, items: Type[CT], name: str = None, *, json_name: str = None, **options) -> None:
        super().__init__(name, json_name=json_name)
        self.items = items
        self.options = AttributeDict(options)

    def __get__(self, instance: 'venom.message.Message', owner):
        if instance is None:
            return self
        return _RepeatValueProxy(instance, self.name)

    def __eq__(self, other):
        if not isinstance(other, RepeatField):
            return False
        return self.items == other.items and self.name == other.name and self.options == other.options

    def __repr__(self):
        return '<{} {} {}>'.format(self.__class__.__name__, self.name, str(self.items))


class MapField(Generic[CT], FieldDescriptor):
    def __init__(self, values: Type[CT], name: str = None, *, json_name: str = None, **options) -> None:
        super().__init__(name, json_name=json_name)
        self.keys = String()
        self.values = values
        self.options = AttributeDict(options)


def Repeat(items: Union[Field, MapField, RepeatField, type, str], **kwargs) -> RepeatField:
    if isinstance(items, type) and issubclass(items, Field):
        items = items()
    if not isinstance(items, (Field, MapField, RepeatField)):
        items = Field(items)
    return RepeatField(items, **kwargs)


def Map(values: Union[Field, MapField, RepeatField, type, str], **kwargs) -> MapField:
    # TODO keys argument.
    return MapField(values, **kwargs)



from typing import Type, Set

from venom import Message
from venom.message import fields
from venom.rpc.method import Method
from venom.fields import RepeatField, MapField, FieldDescriptor
from venom.rpc import Service


class Reflect(object):
    services : Set[Service]
    methods : Set[Type[Method]]
    messages : Set[Type[Message]]

    def __init__(self, title='', version=''):
        self.title = title
        self.version = version
        self.services = set()
        self.methods = set()
        self.messages = set()

    def _add_field(self, field: FieldDescriptor):
        if isinstance(field, RepeatField):
            self._add_field(field.items)
        elif isinstance(field, MapField):
            self._add_field(field.values)
        else:
            if issubclass(field.type, Message):
                self._add_message(field.type)

    def _add_message(self, message: Message):
        if message not in self.messages:
            self.messages.add(message)
            for field in fields(message):
                self._add_field(field)

    def _add_method(self, method: Method):
        self._add_message(method.request)
        self._add_message(method.response)
        self.methods.add(method)

    def add(self, service: Service):
        for method in service.__methods__.values():
            self._add_method(method)
        self.services.add(service)



from typing import Type

import aiohttp
import asyncio

from venom.exceptions import Error, ErrorResponse
from venom.rpc.comms import AbstractClient
from venom.rpc.method import Method, HTTPVerb, HTTPFieldLocation
from venom.protocol import JSON, Protocol, DictProtocol, URIString

try:
    from aiohttp import web, ClientSession
except ImportError:
    raise RuntimeError("You must install the 'aiohttp' package to use the AioHTTP features of Venom RPC")


def _route_handler(venom: 'venom.rpc.Venom',
                   method: Method,
                   protocol_factory: Type[Protocol],
                   query_protocol_factory: Type[DictProtocol] = URIString,
                   path_protocol_factory: Type[DictProtocol] = URIString):
    rpc_response = protocol_factory(method.response)
    rpc_error_response = protocol_factory(ErrorResponse)

    http_status = method.http_status

    http_field_locations = method.http_field_locations()
    http_request_body = protocol_factory(method.request, http_field_locations[HTTPFieldLocation.BODY])
    http_request_query = query_protocol_factory(method.request, http_field_locations[HTTPFieldLocation.QUERY])
    http_request_path = path_protocol_factory(method.request, http_field_locations[HTTPFieldLocation.PATH])

    async def handler(http_request):
        try:
            request = http_request_body.unpack(await http_request.read())
            http_request_query.decode(http_request.url.query, request)
            http_request_path.decode(http_request.match_info, request)

            response = await venom.invoke(method, request)
            return web.Response(body=rpc_response.pack(response),
                                content_type=rpc_response.mime,
                                status=http_status)
        except Error as e:
            return web.Response(body=rpc_error_response.pack(e.format()),
                                content_type=rpc_error_response.mime,
                                status=e.http_status)
    return handler


def create_app(venom: 'venom.rpc.Venom',
               app: web.Application = None,
               protocol_factory: Type[Protocol] = JSON):
    if app is None:
        app = web.Application()

    for method in venom.iter_methods():
        app.router.add_route(method.http_method.value,
                             method.http_path,
                             _route_handler(venom, method, protocol_factory))

    return app


class HTTPClient(AbstractClient):
    def __init__(self,
                 stub: Type['venom.rpc.Service'],
                 base_url: str,
                 *,
                 protocol_factory: Type[Protocol] = None,
                 query_protocol_factory: Type[DictProtocol] = URIString,
                 path_protocol_factory: Type[DictProtocol] = URIString,
                 session: aiohttp.ClientSession = None,
                 **session_kwargs):
        super().__init__(stub, protocol_factory=protocol_factory)
        self._base_url = base_url
        self._query_protocol_factory = query_protocol_factory
        self._path_protocol_factory = path_protocol_factory

        if session is None:
            self._session = aiohttp.ClientSession(**session_kwargs)
        else:
            self._session = session

    async def invoke(self,
                     method: Method,
                     request: 'venom.message.Message',
                     *,
                     context: 'venom.RequestContext' = None,
                     loop: 'asyncio.AbstractEventLoop' = None,
                     timeout: int = None):

        # TODO optional timeouts

        if method.http_path_parameters():
            url = self._base_url + method.http_path.format(**request)
        else:
            url = self._base_url + method.http_path

        headers = None
        if method.http_method in (HTTPVerb.POST, HTTPVerb.PUT, HTTPVerb.PATCH):
            headers = {'content-type': self._protocol_factory.mime}

        http_field_locations = method.http_field_locations()

        # TODO cache this for each RPC call in HTTPClient.__init__()
        params = self._query_protocol_factory(method.request,
                                              http_field_locations[HTTPFieldLocation.QUERY]).encode(request)
        body = self._protocol_factory(method.request,
                                      http_field_locations[HTTPFieldLocation.BODY]).pack(request)

        async with self._session.request(method.http_method.value.lower(), url,
                                         headers=headers,
                                         data=body,
                                         params=params) as response:
            if 200 <= response.status < 400:
                return self._protocol_factory(method.response).unpack(await response.read())
            else:
                self._protocol_factory(ErrorResponse).unpack(await response.read()).raise_()

    # XXX not sure if session should be opened for each request, and why an unclosed session is such a bad thing.
    def __del__(self):
        if self._session:
            self._session.close()


Client = HTTPClient



from abc import ABCMeta
from collections import MutableMapping
from collections import OrderedDict
from typing import Any, Dict, Type, Iterable, TypeVar, Tuple, Set, ClassVar, cast

from venom.fields import FieldDescriptor
from venom.util import meta


class OneOf(object):
    def __init__(self, *choices):
        self.choices = choices

    # TODO helper functions.

    def which(self):
        raise NotImplementedError

    def get(self) -> Any:
        raise NotImplementedError


class MessageMeta(ABCMeta):
    @classmethod
    def __prepare__(metacls, name, bases):
        return OrderedDict()

    def __new__(metacls, name, bases, members):
        cls = super(MessageMeta, metacls).__new__(metacls, name, bases, members)
        cls.__fields__ = OrderedDict(getattr(cls, '__fields__') or ())
        cls.__meta__, meta_changes = meta(bases, members)
        cls.__meta__.protocols = {}

        if not meta_changes.get('name', None):
            cls.__meta__.name = name

        for name, member in members.items():
            if isinstance(member, FieldDescriptor):
                cls.__fields__[member.name] = member
            elif isinstance(member, OneOf):
                cls.__meta__.one_of_groups += (name, member.choices)

        return cls


class Message(MutableMapping, metaclass=MessageMeta):
    __slots__ = ('_values',)   # TODO slot message fields directly.
    # TODO change to tuple (FieldDescriptor would need FieldDescriptor.attribute attribute.)
    __fields__: ClassVar[Dict[str, FieldDescriptor]] = None
    __meta__: ClassVar[Dict[str, Any]] = None

    class Meta:
        name = None
        one_of_groups = ()
        protocols = None

    def __init__(self, *args, **kwargs):
        if args:
            self._values = {}
            for value, key in zip(args, self.__fields__.keys()):
                self[key] = value
            for key, value in kwargs.items():
                self[key] = value
        else:
            self._values = {key: value for key, value in kwargs.items() if value is not None}

    def get(self, key, default=None):
        try:
            return self._values[key]
        except KeyError:
            if default is None and key in self.__fields__:
                return self.__fields__[key].default()
            return default

    def __getitem__(self, key):
        return self._values[key]

    def __setitem__(self, key, value):
        if value is None:
            try:
                del self._values[key]
            except KeyError:
                pass
        else:
            self._values[key] = value

    def __delitem__(self, key):
        del self._values[key]

    def __contains__(self, key):
        return key in self._values

    def __iter__(self):
        return iter(self._values)

    def __len__(self):
        return len(self._values)

    def __repr__(self):
        parts = []
        for key in self.__fields__.keys():
            if key in self._values:
                parts.append('{}={}'.format(key, repr(self._values[key])))
        return '{}({})'.format(self.__meta__.name, ', '.join(parts))


def fields(message: Type[Message]) -> Iterable[FieldDescriptor]:
    return tuple(message.__fields__.values())


def field_names(message: Type[Message]) -> Iterable[FieldDescriptor]:
    return tuple(field.name for field in message.__fields__.values())


def is_empty(message: Type[Message]) -> bool:
    return not fields(message)


_M = TypeVar('M', bound=Message)


def from_object(message: Type[_M], obj: Any) -> _M:
    kwargs = {}

    for name in message.__fields__.keys():
        if hasattr(obj, '__getitem__'):
            # TODO skip None values
            try:
                kwargs[name] = obj[name]
                continue
            except (IndexError, TypeError, KeyError):
                pass
        try:
            kwargs[name] = getattr(obj, name)
        except AttributeError:
            pass

    return message(**kwargs)


def one_of(*choices):
    """

    Usage:::

        class SearchRequest(Message):
            query = one_of('name', 'id')

        s = SearchRequest(id=123)
        s.query.which()  # 'id'


    """
    return OneOf(choices)


class Empty(Message):
    pass


def message_factory(name: str, fields: Dict[str, FieldDescriptor], *, super_message: Type[_M] = Message) -> Type[_M]:
    return cast(Type[_M], type(name, (super_message,), fields))



from abc import ABCMeta, abstractmethod
from base64 import b64encode, b64decode
from functools import partial
from itertools import chain
from json import JSONDecodeError
from typing import Type, TypeVar, Generic, Callable, Union, Any, Tuple, Iterable, Dict, List, Set, MutableMapping, \
    Mapping

from venom import Empty
from venom import Message
from venom.exceptions import ValidationError
from venom.fields import Field, ConverterField, RepeatField, FieldDescriptor, MapField
from venom.message import field_names, fields


# def partial(message: Type[Message], include: Set[str], name: str = None):
#     if len(include) == 0:
#         return Empty
#     elif set(field_names(message)).issuperset(include):
#         return message
#     else:
#         if name is None:
#             name = 'Partial{}'.format(message.__meta__.name)
#
#         return message_factory(name, {name: field for name, field in fields(message) if name in include})


class Protocol(metaclass=ABCMeta):
    mime: str = None
    name: str = None

    def __new__(cls, fmt: Type[Message], field_names_: Set[str] = None):
        if field_names_ is None:
            try:
                return fmt.__meta__.protocols[cls.name]
            except KeyError:
                instance = super(Protocol, cls).__new__(cls)
                instance.__init__(fmt)
                return instance
        # elif len(field_names) == 0:
        #     return Protocol.__new__(cls, Empty)
        else:
            instance = super(Protocol, cls).__new__(cls)
            instance.__init__(fmt, field_names_)
            return instance

    def __init__(self, fmt: Type[Message], field_names_: Set[str] = None):
        if field_names_ is None:
            field_names_ = field_names(fmt)
            fmt.__meta__.protocols[self.name] = self
        self._format = fmt
        self._fields = [field for field in fields(fmt) if field.name in field_names_]

    @classmethod
    def _get_protocol(cls, fmt: Type[Message]):
        try:
            return fmt.__meta__.protocols[cls.name]
        except KeyError:
            return cls(fmt)

    @abstractmethod
    def pack(self, message: Message) -> bytes:
        pass

    @abstractmethod
    def unpack(self, buffer: bytes) -> Message:
        pass


try:
    import ujson as json
except ImportError:
    import json

JSONPrimitive = Union[str, int, float, bool]

JSONValue = Union[JSONPrimitive, Dict[str, JSONPrimitive], List[JSONPrimitive]]


class DictProtocol(Protocol, metaclass=ABCMeta):
    @abstractmethod
    def encode(self, message: Message):
        pass

    @abstractmethod
    def decode(self, instance: Any, message: Message = None) -> Message:
        pass


class JSON(DictProtocol):
    mime = 'application/json'
    name = 'json'

    def __init__(self, fmt: Type[Message], field_names_: Set[str] = None):
        super().__init__(fmt, field_names_)
        # TODO camelCase conversion
        self.field_encoders = {(field.name, field.json_name): self._field_encoder(field) for field in self._fields}
        self.field_decoders = {(field.name, field.json_name): self._field_decoder(field) for field in self._fields}

    T = TypeVar('T')

    @staticmethod
    def _cast(type_: Type[T], value: Any) -> T:
        # TODO JSON type names, i.e. object instead of dict, integer instead of int etc.
        if not isinstance(value, type_):
            raise ValidationError(f"{repr(value)} is not of type '{type_.__name__}'")
        return value

    @staticmethod
    def _cast_number(value: Any):
        if type(value) not in (int, float):
            raise ValidationError(f"{value} is not a number")
        return float(value)

    def _field_encoder(self, field: FieldDescriptor) -> Callable[[Any], JSONValue]:
        if isinstance(field, RepeatField):
            field_item_encoder = self._field_encoder(field.items)
            return lambda lst: [field_item_encoder(item) for item in lst]

        if isinstance(field, MapField):
            field_item_encoder = self._field_encoder(field.values)
            return lambda dct: {k: field_item_encoder(v) for k, v in dct.items()}  # TODO: keys?

        if not isinstance(field, Field):
            raise NotImplementedError()

        if issubclass(field.type, Message):
            field_protocol = self._get_protocol(field.type)
            return lambda msg: field_protocol.encode(msg)

        if field.type is bytes:
            return lambda b: b64encode(b)

        # assume all is JSON from here
        return lambda value: value

    def _field_decoder(self, field: FieldDescriptor) -> Callable[[JSONValue], Any]:
        if isinstance(field, RepeatField):
            field_item_decoder = self._field_decoder(field.items)
            return lambda lst: [field_item_decoder(item) for item in self._cast(list, lst)]

        if isinstance(field, MapField):
            field_item_decoder = self._field_decoder(field.values)
            return lambda dct: {k: field_item_decoder(v) for k, v in dct.items()}  # TODO: keys?

        if not isinstance(field, Field):
            raise NotImplementedError()

        if issubclass(field.type, Message):
            field_protocol = self._get_protocol(field.type)
            return lambda msg: field_protocol.decode(msg)

        # an integer (int) in JSON is also a number (float), so we convert here if necessary:
        if field.type is float:
            return self._cast_number

        if field.type is bytes:
            # TODO catch TypeError
            return lambda b: b64decode(b)

        return partial(self._cast, field.type)

    def encode(self, message: Message):
        obj = {}
        for (name, json_name), encode in self.field_encoders.items():
            try:
                obj[json_name] = encode(message[name])
            except KeyError:
                pass
        return obj

    def decode(self, instance: Any, message: Message = None) -> Message:
        if not isinstance(instance, Mapping):
            raise ValidationError(f"{repr(instance)} is not of type 'object'")

        if message is None:
            message = self._format()

        for (name, json_name), decode in self.field_decoders.items():
            if json_name in instance:
                try:
                    message[name] = decode(instance[json_name])
                except ValidationError as e:
                    e.path.insert(0, json_name)
                    raise e
        return message

    # TODO include It
    def pack(self, message: Message, include: Iterable[str] = None) -> bytes:
        if self._format is Empty:
            return b''
        return json.dumps(self.encode(message)).encode('utf-8')

    def unpack(self, buffer: bytes):
        # Allow empty string when message is empty
        if len(buffer) == 0 and not self._fields:
            return self._format()

        try:
            return self.decode(json.loads(buffer.decode('utf-8')))
        except (ValueError, JSONDecodeError) as e:
            raise ValidationError(f"Invalid JSON: {str(e)}")


class URIString(JSON):

    def _field_decoder(self, field: FieldDescriptor) -> Callable[[JSONValue], Any]:
        if isinstance(field, RepeatField):
            raise NotImplementedError(f'Unable to decode {field} from URI string')

        if not isinstance(field, Field):
            raise NotImplementedError()

        if issubclass(field.type, Message):
            raise NotImplementedError(f'Unable to decode {field} from URI string')

        if field.type is str:
            return lambda s: s

        if field.type is bytes:
            # TODO catch TypeError
            return lambda b: b64decode(b)

        return partial(self._cast, field.type)

    def _field_encoder(self, field: FieldDescriptor) -> Callable[[Any], JSONValue]:
        if isinstance(field, RepeatField):
            raise NotImplementedError(f'Unable to decode {field} from URI string')

        if not isinstance(field, Field):
            raise NotImplementedError()

        if issubclass(field.type, Message):
            raise NotImplementedError(f'Unable to decode {field} from URI string')

        if field.type is bytes:
            return lambda b: b64encode(b)

        return lambda x: str(x)

    @staticmethod
    def _cast(type_: type, value: Any):
        # TODO JSON/wire-format specific type names, i.e. object instead of dict, integer instead of int etc.
        try:
            return type_(value)
        except ValueError:
            raise ValidationError(f"{repr(value)} is not formatted as a '{type_.__name__}'")



import asyncio
import unittest
from functools import wraps
from typing import Union, Type, Iterable
from unittest.mock import MagicMock
from weakref import WeakKeyDictionary
from venom.rpc import Venom, UnknownService, Service
from venom.rpc.context import RequestContext


class AsyncMock(MagicMock):
    async def __call__(self, *args, **kwargs):
        return super(AsyncMock, self).__call__(*args, **kwargs)


class MockVenom(Venom):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._context_mock_instances = WeakKeyDictionary()

    def get_instance(self, reference: Union[str, type]):
        context = RequestContext.current()

        if context is not None:
            try:
                return self._context_mock_instances[context][reference]
            except KeyError:
                pass

        try:
            return super().get_instance(reference)
        except UnknownService:
            instance = AsyncMock()

        if context is not None:
            if context in self._context_mock_instances:
                self._context_mock_instances[context][reference] = instance
            else:
                self._context_mock_instances[context] = {reference: instance}

        return instance


def mock_venom(*services: Iterable[Type[Service]], **kwargs):
    venom = MockVenom(**kwargs)
    for service in services:
        venom.add(service)
    return venom


def mock_instance(service: Type[Service], *dependencies: Iterable[Type[Service]], **kwargs):
    venom = mock_venom(service, *dependencies, **kwargs)
    return venom.get_instance(service)


def sync(coro):
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    future = loop.create_task(coro)
    loop.run_until_complete(future)
    return future.result()


def sync_decorator(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        return sync(fn(*args, **kwargs))

    return wrapper


class AioTestCaseMeta(type):
    def __new__(mcs, name, bases, members):
        for key, value in members.items():
            if key.startswith('test_') and asyncio.iscoroutinefunction(value):
                members[key] = sync_decorator(value)

        return super(AioTestCaseMeta, mcs).__new__(mcs, name, bases, members)


class AioTestCase(unittest.TestCase, metaclass=AioTestCaseMeta):
    """
    A custom unittest.TestCase that converts all tests that are coroutine functions into synchronous tests.
    """
    pass



from venom.rpc import Service, Venom
from venom.rpc import http
from venom.rpc.reflect.openapi import make_openapi_schema
from venom.rpc.reflect.reflect import Reflect
from venom.rpc.reflect.stubs import ReflectStub, OpenAPISchema


class ReflectService(Service):
    class Meta:
        stub = ReflectStub

    def __init__(self, venom: 'venom.rpc.Venom' = None) -> None:
        super().__init__(venom)
        self.reflect = Reflect(
            title=venom.options.get('title', 'API'),
            version=venom.options.get('version', '1.0.0'),
        )

        for service in venom:
            self._reflect_service(venom, service)

        Venom.on_add_public_service.connect(self._reflect_service, sender=venom)

    def _reflect_service(self, sender: Venom, service: Service):
        self.reflect.add(service)

    @http.GET('/openapi.json')
    def get_openapi_schema(self) -> OpenAPISchema:
        return make_openapi_schema(self.reflect)





from typing import Union, List, Mapping, NewType

_JSONArray = List['_JSONValue']
_JSONObject = Mapping[str, '_JSONValue']
_JSONValue = Union[None, bool, str, int, float, '_JSONObject', '_JSONArray']


JSONArray = NewType('JSONArray', _JSONArray)
JSONObject = NewType('JSONObject', _JSONObject)
JSONValue = NewType('JSONValue', _JSONValue)



import enum

from typing import Iterable, Any, Tuple

from venom.fields import String, Int32, Int64, Bool, Float32, Float64, Repeat, Bytes, Field, Map
from venom.message import Message, one_of


class StringValue(Message):
    value = String()

    class Meta:
        proto_package = 'google.protobuf'


class BytesValue(Message):
    value = Bytes()

    class Meta:
        proto_package = 'google.protobuf'


class BoolValue(Message):
    value = Bool()

    class Meta:
        proto_package = 'google.protobuf'


class Int32Value(Message):
    value = Int32()

    class Meta:
        proto_package = 'google.protobuf'


class Int64Value(Message):
    value = Int64()

    class Meta:
        proto_package = 'google.protobuf'


IntegerValue = Int64Value


class Float32Value(Message):
    value = Float32()

    class Meta:
        proto_package = 'google.protobuf'


class Float64Value(Message):
    value = Float64()

    class Meta:
        proto_package = 'google.protobuf'


NumberValue = Float64Value


class NullValue(enum.Enum):
    NULL_VALUE = None


class Struct(Message):
    fields = Map('venom.common.messages.Value')


class ListValue(Message):
    values = Repeat('venom.common.messages.Value')


class Value(Message):
    # null_value = Field(NullValue)
    number_value = Field(NumberValue)
    string_value = Field(StringValue)
    bool_value = Field(BoolValue)
    struct_value = Field(Struct)
    list_value = Field(ListValue)

    value = one_of(  # null_value,
                   number_value,
                   string_value,
                   bool_value,
                   struct_value,
                   list_value)


from venom.protocol import JSON as _JSON


class _JSONValue(_JSON):
    def encode(self, message: Message):
        return message.value

    def decode(self, instance: Any, skip: Iterable[str] = ()) -> Message:
        value = self._cast(self._format.value.type, instance)
        return self._format(value)


StringValue.__meta__.protocols['json'] = _JSONValue(StringValue)
IntegerValue.__meta__.protocols['json'] = _JSONValue(IntegerValue)
NumberValue.__meta__.protocols['json'] = _JSONValue(NumberValue)
BoolValue.__meta__.protocols['json'] = _JSONValue(BoolValue)


class FieldMask(Message):
    paths = Repeat(String())

    class Meta:
        proto_package = 'google.protobuf'

    # TODO needs tests
    def match_path(self, *path: Tuple[str]) -> bool:
        for level in range(len(path)):
            match_path = '.'.join(path[:level + 1])
            for path_ in self.paths:
                if match_path in path_:
                    return True
        return False


class _JSONFieldMask(_JSON):
    def encode(self, message: FieldMask):
        return ','.join(message.paths)

    def decode(self, instance: Any, skip: Iterable[str] = ()) -> Message:
        paths = self._cast(str, instance)
        return self._format(paths.split(','))


FieldMask.__meta__.protocols['json'] = _JSONFieldMask(FieldMask)


class Timestamp(Message):
    seconds = Int64()
    nanos = Int32()

    class Meta:
        proto_package = 'google.protobuf'



import asyncio
from functools import wraps
from inspect import signature, Parameter
from typing import Callable, Any, Sequence, get_type_hints, Type, NamedTuple, Optional, List, Dict, TypeVar
from typing import Tuple
from typing import Union
from venom.fields import RepeatField, MapField, Field, ConverterField, Repeat

from venom.converter import Converter

from venom.message import Empty, Message, message_factory, fields, field_names
from venom.rpc.resolver import Resolver
from venom.util import upper_camelcase

MessageFunction = NamedTuple('MessageFunction', [
    ('request', Type[Message]),
    ('response', Type[Message]),
    ('invokable', Callable[[Any, Message, Optional['asyncio.AbstractEventLoop']], Message])  # TODO update to typing.Coroutine in Python 3.6
])


def get_field_type(field):
    # TODO support other mappings and sequences
    # XXX support for list-within-a-list situations?
    if isinstance(field, RepeatField):
        return List[field.items.type]
    elif isinstance(field, MapField):
        return Dict[str, field.values.type]
    elif isinstance(field, Field):
        return field.type
    else:
        raise NotImplementedError


def create_field_from_type(type_, converters: Sequence[Converter] = (), default: Any = None):
    if type_ in (bool, int, float, str, bytes):
        return Field(type_, default=default)

    for converter in converters:
        if converter.python == type_:
            return ConverterField(converter)

    # TODO type_ != Any is a workaround for https://github.com/python/typing/issues/345
    if type_ != Any and issubclass(type_, List):
        # FIXME List[List[X]] must not become Repeat(Repeat(X))
        return Repeat(create_field_from_type(type_.__args__[0]))

    # TODO type_ != Any is a workaround for https://github.com/python/typing/issues/345
    if type_ != Any and issubclass(type_, Message):
        return Field(type_)

    raise NotImplementedError(f"Unable to generate field for {type_}")


def dynamic(name: str, expression: Union[type, Callable[[Type[Any]], type]]) \
        -> Callable[[Callable[..., Any]], Callable[..., Any]]:  # TODO type annotations for pass-through decorator
    """
    
    :param name: 
    :param expression: a subclass of ``type`` or a callable in the format ``(owner: Type[Any]) -> type``.
    :return: 
    """
    def decorator(func):
        if not hasattr(func, '__dynamic__'):
            func.__dynamic__ = {name: expression}
        else:
            func.__dynamic__[name] = expression
        return func
    return decorator


def _get_func_type_annotations(func: Callable[..., Any], owner: type = None) -> Dict[str, type]:
    annotations = dict(get_type_hints(func))
    dynamic_annotations = getattr(func, '__dynamic__', {})

    for name, expression in dynamic_annotations.items():
        if type(expression) == type:
            annotations[name] = expression
        elif callable(expression):
            annotations[name] = expression(owner)
    return annotations


# TODO name arg for use with auto-generation
def magic_normalize(func: Callable[..., Any],
                    func_name: str = None,
                    request: Type[Message] = None,
                    response: Type[Message] = None,
                    *,
                    converters: Sequence[Union[Converter, Type[Converter]]] = (),
                    additional_args: Sequence[Union[Resolver, Type[Resolver]]] = (),
                    owner: type = None,
                    auto_generate_request: bool = False) -> MessageFunction:
    """

    :param func:
    :param request:
    :param response:
    :param converters:
    :param owner: 
    :param additional_args: additional arguments that are resolved during invocation.
    :return:
    """
    if func_name is None:
        func_name = func.__name__

    # TODO parameters supplied by the service implementation through a context; session etc.

    additional_args = [resolver() if isinstance(resolver, type) else resolver for resolver in additional_args]
    converters = [converter() if isinstance(converter, type) else converter for converter in converters]

    func_signature = signature(func)
    func_type_hints = _get_func_type_annotations(func, owner)


    if len(func_signature.parameters) == 0:
        # no "self" parameter: func()
        raise RuntimeError(f"At least one argument expected in {func}")

    request_converter = None
    func_parameters = tuple(func_signature.parameters.items())[1 + len(additional_args):]

    if len(func_parameters):
        name, param = func_parameters[0]
        param_type = func_type_hints.get(name, Any)

        unpack_request: Union[bool, Tuple[str, ...]] = False

        # TODO param_type != Any is a workaround for https://github.com/python/typing/issues/345
        if param_type != Any and issubclass(param_type, Message) and name == 'request':
            # func(self, request: MessageType, ...)
            if request is None:
                request = param_type
            elif request != param_type and 'request' not in field_names(request):
                raise RuntimeError(f"Bad argument in {func}: "
                                   f"'{name}' should be {request}, but got {param_type}")

            for name, param in func_parameters[1:]:
                if param.default is Parameter.empty:
                    raise RuntimeError(f"Unexpected required argument in {func}: '{name}'")
        elif request is not None and name == 'request':
            for converter in converters:
                if converter.wire == request and converter.python == param_type:
                    request_converter = converter
                    break

            if not request_converter:
                raise RuntimeError(f"Unable to coerce request message to python format: "
                                   f"'{param_type}' in {func}")
        else:  # func(self, arg: ?, ...)
            required_params, remaining_params = {}, {}
            for name, param in func_parameters:

                param_type = func_type_hints.get(name, Any)
                if hasattr(param_type, '__supertype__'):  # handles NewType
                    param_type = param_type.__supertype__

                if param.default is Parameter.empty:
                    required_params[name] = param_type
                else:
                    remaining_params[name] = (param_type, param.default)

            if request:  # unpack from request message
                request_fields = request.__fields__
                message_params = set()

                for name, param_type in required_params.items():
                    try:
                        field = request_fields[name]
                    except KeyError:
                        raise RuntimeError(f"Unexpected required argument in {func}: "
                                           f"'{name}' is not a field of {request}")

                    field_type = get_field_type(field)

                    if param_type not in (Any, field_type):
                        raise RuntimeError(f"Bad argument in {func}: "
                                           f"'{name}' should be {field_type}, but got {param_type}")

                    message_params.add((name, None))

                for name, (param_type, default) in remaining_params.items():
                    if name in request_fields:
                        field = request_fields[name]
                        field_type = get_field_type(field)

                        if param_type not in (Any, field_type):
                            raise RuntimeError(f"Bad argument in {func}: "
                                               f"'{name}' should be {field_type}, but got {param_type}")

                        message_params.add((name, default))

                if message_params == field_names(request):
                    unpack_request = True
                else:
                    unpack_request = tuple(message_params)

            else:  # auto-generate message from params (uses converters where necessary)
                unpack_request = True

                if not auto_generate_request:
                    raise RuntimeError(f"Message auto-generation required in {func}")

                message_fields = {}
                for name, param in func_parameters:
                    param_type = func_type_hints.get(name, Any)
                    if hasattr(param_type, '__supertype__'):  # handles NewType
                        param_type = param_type.__supertype__

                    if param.default is Parameter.empty:
                        message_fields[name] = create_field_from_type(param_type, converters=converters)
                    else:
                        message_fields[name] = create_field_from_type(param_type,
                                                                      converters=converters,
                                                                      default=param.default)

                request = message_factory(f'{upper_camelcase(func_name)}Request', message_fields)
    else:  # func(self)
        if request is None:
            request = Empty

        # either request is Empty, or otherwise all request fields are discarded -- strange, but valid
        unpack_request = ()

    return_type = func_type_hints.get('return', Any)
    response_converter = None

    if response is None:
        if hasattr(return_type, '__supertype__'):  # handles NewType
            return_type = return_type.__supertype__

        if return_type in (Any, None, type(None)):  # None for Python 3.5 compatibility
            # TODO warn if Any: missing return type annotation (will discard return value)
            response = Empty
        elif issubclass(return_type, Message):
            response = return_type
        else:
            for converter in converters:
                if converter.python == return_type:
                    response_converter = converter
                    response = converter.wire
                    break

            if not response_converter:
                raise RuntimeError(f"Unable to coerce return value to wire format: "
                                   f"'{return_type}' in {func}")
    elif return_type == Any:
        # NOTE missing return type annotation (assuming 'response' here)
        pass
    elif return_type in (None, type(None)):  # None for Python 3.5 compatibility
        response = Empty
    elif response != return_type:
        for converter in converters:
            if converter.wire == response and converter.python == return_type:
                response_converter = converter
                break

        if not response_converter:
            raise RuntimeError(f"Unable to coerce return value to wire format: "
                               f"'{return_type}' in {func}")

    wrap_args = lambda req: (req,)
    wrap_kwargs = lambda req: {}

    if unpack_request is True:
        wrap_args = lambda req: ()
        wrap_kwargs = lambda req: req
    elif unpack_request is False:
        if request_converter:
            wrap_args = lambda req: (request_converter.convert(req),)
    elif unpack_request == ():
        wrap_args = lambda req: ()
    else:
        wrap_args = lambda req: ()
        wrap_kwargs = lambda req: {f: req.get(f, d) for f, d in unpack_request}

    if response == Empty and return_type != Empty:
        wrap_response = lambda res: Empty()
    elif response_converter:
        wrap_response = lambda res: response_converter.format(res)
    else:
        wrap_response = lambda res: res

    if not asyncio.iscoroutinefunction(func):
        func_ = func

        @wraps(func)
        async def func(*args, **kwargs):
            return func_(*args, **kwargs)

    # TODO (optimization) do not wrap what does not need to be wrapped
    if additional_args:
        async def invokable(inst, req: Message, loop: 'asyncio.AbstractEventLoop' = None) -> Message:
            return wrap_response(await func(inst,
                                            *(await asyncio.gather(*[arg.resolve(inst, req)
                                                                     for arg in additional_args], loop=loop)),
                                            *wrap_args(req),
                                            **wrap_kwargs(req)))
    else:
        async def invokable(inst, req: Message, loop: 'asyncio.AbstractEventLoop' = None) -> Message:
            return wrap_response(await func(inst, *wrap_args(req), **wrap_kwargs(req)))

    return MessageFunction(request, response, wraps(func)(invokable))



from typing import Iterable
from itertools import groupby, chain
from functools import singledispatch
from collections import defaultdict
from operator import attrgetter
from venom.rpc.reflect.reflect import Reflect
from venom.rpc.reflect.stubs import OperationMessage, \
    ParameterMessage, ResponsesMessage, ResponseMessage, \
    SchemaMessage, InfoMessage, OpenAPISchema
from venom.protocol import JSON
from venom.fields import Field, RepeatField, MapField
from venom import Message
from venom.message import fields, is_empty
from venom.rpc.method import Method, HTTPFieldLocation


DESCRIPTION = 'description'


# TODO: other types
TYPE_TO_JSON = {
    str: 'string',
    int: 'integer',
    float: 'double',
    bool: 'boolean',
}

PATH_PARAMETER = {
    'is_in': 'path',
    'required': True
}

BODY_PARAMETER = {
    'is_in': 'body',
}

QUERY_PARAMETER = {
    'is_in': 'query',
}


def ref_schema_message(message: Message) -> SchemaMessage:
    return SchemaMessage(ref=f'#/definitions/{message.__meta__.name}')


@singledispatch
def schema_message(field: Field) -> SchemaMessage:
    if field.type not in TYPE_TO_JSON:
        if issubclass(field.type, Message):
            return ref_schema_message(field.type)
        return schema_message(field.type)
    return SchemaMessage(
        type=TYPE_TO_JSON[field.type],
        description=field.options.get(DESCRIPTION),
    )


@schema_message.register(RepeatField)
def schema_message_repeat(field: RepeatField) -> SchemaMessage:
    return SchemaMessage(
        type='array',
        items_=schema_message(field.items),
        description=field.options.get(DESCRIPTION),
    )


@schema_message.register(MapField)
def schema_message_map(field: MapField) -> SchemaMessage:
    return SchemaMessage(
        type='object',
        additionalProperties=schema_message(field.values),
        description=field.options.get(DESCRIPTION),
    )


def parameters_at_location(request: Message, names: Iterable, default: dict):
    for field_name in names:
        field = getattr(request, field_name)
        field_schema = schema_message(field)
        protocol = JSON(type(field_schema))
        yield ParameterMessage(
            name=field.json_name,
            **default,
            **protocol.encode(field_schema),
        )


def parameters_path(method: Method) -> Iterable:
    return parameters_at_location(
        method.request,
        method.http_path_parameters(),
        PATH_PARAMETER
    )

def parameters_query(method: Method) -> Iterable:
    return parameters_at_location(
        method.request,
        method.http_field_locations()[HTTPFieldLocation.QUERY],
        QUERY_PARAMETER
    )

def parameters_body(method: Method) -> list:
    body_fields = method.http_field_locations()[HTTPFieldLocation.BODY]
    if not body_fields:
        return []
    fields = {f: getattr(method.request, f) for f in body_fields}
    if fields == method.request.__fields__:
        param = dict(
            name=method.request.__meta__.name,
            schema=ref_schema_message(method.response)
        )
    else:
        param = dict(
            name=method.name + '_body',
            schema=schema_for_fields(fields.values())
        )
    return [ParameterMessage(**BODY_PARAMETER, **param)]


def response_message(method: Method) -> ResponseMessage:
    return ResponseMessage(
        description=method.options.get(DESCRIPTION, ''),
        schema=ref_schema_message(method.response)
    )


def operation_message(method: Method) -> OperationMessage:
    return OperationMessage(
        responses=ResponsesMessage(default=response_message(method)),
        parameters=list(chain(
            parameters_body(method),
            parameters_path(method),
            parameters_query(method)
        ))
    )


def schema_for_fields(fields, description=None) -> SchemaMessage:
    return SchemaMessage(
        type='object',
        properties={
            v.json_name: schema_message(v) for v in fields
            },
        description=description,
    )


def schema_for_message(message: Message) -> SchemaMessage:
    return schema_for_fields(
        fields(message),
        description=message.__meta__.get(DESCRIPTION)
    )


def make_openapi_schema(reflect: Reflect) -> OpenAPISchema:
    paths = defaultdict(dict)
    for path, group in groupby(reflect.methods, key=attrgetter('http_path')):
        for method in group:
            paths[path.strip('.')][method.http_method.value.lower()] = operation_message(method)
    definitions = {
        m.__meta__.name: schema_for_message(m) for m in reflect.messages
        if not is_empty(m)
    }
    return OpenAPISchema(
        swagger='2.0',
        consumes=['application/json'],
        produces=['application/json'],
        info=InfoMessage(version=reflect.version, title=reflect.title),
        paths=dict(paths),
        definitions=definitions,
    )



import calendar
from typing import TypeVar, Generic

import datetime

from venom.common import types
from venom.common.messages import Int64Value, Int32Value, StringValue, Float32Value, Float64Value, BoolValue, Timestamp, \
    BytesValue, Value
from venom.converter import Converter

V = TypeVar('V')
T = TypeVar('T')


class _ValueConverter(Converter[V, T]):
    def convert(self, message: V) -> T:
        return message.value

    def format(self, value: T) -> V:
        return self.wire(value=value)


class StringValueConverter(_ValueConverter[StringValue, str]):
    wire = StringValue
    python = str


class BytesValueConverter(_ValueConverter[BytesValue, bytes]):
    wire = BytesValue
    python = bytes


class BooleanValueConverter(_ValueConverter[StringValue, bool]):
    wire = BoolValue
    python = bool


class Int32ValueConverter(_ValueConverter[Int32Value, int]):
    wire = Int32Value
    python = int


class Int64ValueConverter(_ValueConverter[Int64Value, int]):
    wire = Int64Value
    python = int


IntegerValueConverter = Int64ValueConverter


class Float32ValueConverter(_ValueConverter[Float32Value, float]):
    wire = Float32Value
    python = float


class Float64ValueConverter(_ValueConverter[Float64Value, float]):
    wire = Float64Value
    python = float


NumberValueConverter = Float64ValueConverter


class DateConverter(Converter):
    wire = Timestamp
    python = datetime.date

    def convert(self, value: Timestamp) -> datetime.date:
        return datetime.datetime.fromtimestamp(value.seconds).date()

    def format(self, value: datetime.date) -> Timestamp:
        return Timestamp(int(calendar.timegm(value.timetuple())))


class DateTimeConverter(Converter):
    wire = Timestamp
    python = datetime.datetime

    def convert(self, value: Timestamp) -> datetime.datetime:
        return datetime.datetime.utcfromtimestamp(value.seconds + value.nanos / 10 ** 9)

    def format(self, value: datetime.datetime) -> Timestamp:
        unix = calendar.timegm(value.utctimetuple())
        seconds = int(unix)
        nanos = int((unix - seconds) * 10 ** 9)
        return Timestamp(seconds, nanos)


# TODO
# class JSONValueConverter(Converter):
#     wire = Value
#     python = types.JSONValue
#
#     @staticmethod
#     def _convert_number_value(value: Value):
#         return value.number_value
#
#     def convert(self, value: Value) -> types.JSONValue:
#         return {
#             'number_value': self._convert_number_value
#         }[value.value.which()](value)



from typing import Type

from venom.rpc.method import MethodDescriptor
from venom.rpc.service import Service, ServiceMeta, ServiceManager


class StubManager(ServiceManager):

    def prepare_method(self, service: Type['Service'], method: MethodDescriptor, name: str):
        return method.stub(service, name)


class StubMeta(ServiceMeta):
    def __new__(metacls, what, bases=None, members=None):
        cls = super().__new__(metacls, what, bases, members)
        cls.__method_descriptors__ = {
            name: method.stub(cls, name)
            for name, method in cls.__method_descriptors__.items()
        }
        return cls


class Stub(Service, metaclass=StubMeta):
    class Meta:
        manager = StubManager



from abc import abstractmethod, ABC
from types import MethodType
from typing import Type, Generic, TypeVar, ClassVar, Any

from venom.exceptions import NotImplemented_
from venom.protocol import Protocol, JSON
from venom.rpc.method import Method, Req, Res

S = TypeVar('S', bound='venom.rcp.service.Service')


class ClientMethod(Method[S, Req, Res]):
    def __init__(self, client: 'AbstractClient[S]', stub_method: Method[S, Req, Res]):
        super().__init__(stub_method.name,
                         stub_method.request,
                         stub_method.response,
                         service=stub_method.service,
                         http_path=stub_method.http_path,
                         http_method=stub_method.http_method,
                         http_status=stub_method.http_status,
                         **stub_method.options)
        self.client = client

    async def invoke(self, instance, request: Req, loop: 'asyncio.AbstractEventLoop' = None) -> Res:
        return await self.client.invoke(self, request, loop=loop)


class AbstractClient(ABC, Generic[S]):
    client_method_cls: ClassVar[Type['ClientMethod']] = ClientMethod
    stub: S

    def __init__(self, stub: S, *, protocol_factory: Type[Protocol] = None):
        self.stub = stub
        # NOTE method bindings for all methods in the stub.
        for name, stub_method in stub.__methods__.items():
            method = self.client_method_cls(self, stub_method)
            # self.__methods__[name] = method
            setattr(self, name, method.__get__(stub))

        if protocol_factory is None:
            protocol_factory = JSON

        self._protocol_factory = protocol_factory

    @abstractmethod
    async def invoke(self, method: Method, request: 'venom.message.Message', *, loop: 'asyncio.AbstractEventLoop' = None):
        raise NotImplementedError



import asyncio
from typing import Optional, MutableMapping

from venom.rpc.resolver import Resolver
from weakref import WeakKeyDictionary


class RequestContext(object):
    __contexts: MutableMapping[asyncio.Task, 'RequestContext'] = WeakKeyDictionary()
    _context_task: asyncio.Task = None

    def __init__(self, venom: 'venom.rpc.Venom' = None):
        self.venom = venom

    @classmethod
    def current(cls) -> Optional['RequestContext']:
        current_task = asyncio.Task.current_task()
        if current_task is None:
            return None
        return cls.__contexts.get(current_task)

    def __enter__(self) -> 'RequestContext':
        current_task = asyncio.Task.current_task()

        if current_task is None:
            raise RuntimeError('Unable to create RequestContext: No current task')

        if self._context_task is not None:
            raise RuntimeError('Unable to re-enter RequestContext: This context has already been entered')

        self._context_task = current_task
        self.__contexts[current_task] = self
        return self

    def __exit__(self, *args) -> None:
        del self.__contexts[self._context_task]


class RequestContextResolver(Resolver):
    python = RequestContext

    async def resolve(self,
                      service: 'venom.rpc.service.Service',
                      request: 'venom.message.Message') -> RequestContext:
        return service.context


class RequestContextDescriptor(object):
    def __get__(self, instance, owner) -> Optional['RequestContext']:
        return RequestContext.current()


class DictRequestContext(RequestContext, dict):
    def __hash__(self):
        return hash(self._context_task)



from venom import Message
from venom.fields import Int32, String


# TODO consider moving that into the implementations. Should at least be moved into rpc.*
class ErrorResponse(Message):
    status = Int32()
    description = String()

    # TODO Repeat(String)
    path = String()

    # TODO helper for raising errors
    def raise_(self):
        if self.status == 501:
            raise NotImplemented_()
        # if self.status == 500:
        #         raise ServerError(self.message or '')
        #     if self.status == 404:
        #         raise NotFound(self.message or '')

        raise RuntimeError('HTTP status {}: {}'.format(self.status, self.description))


class Error(Exception):
    http_status = 500
    description = ''

    def __init__(self, message: str = None):
        if message:
            self.description = message

    def format(self) -> ErrorResponse:
        return ErrorResponse(status=self.http_status, description=self.description)


class NotImplemented_(Error):
    http_status = 501
    description = 'Not Implemented'


class NotFound(Error):
    http_status = 404
    description = 'Not Found'


class BadRequest(Error):
    http_status = 400
    description = 'Bad Request'


class Unauthorized(Error):
    http_status = 401
    description = 'Unauthorized'


class Forbidden(Error):
    http_status = 403
    description = 'Forbidden'


class Conflict(Error):
    http_status = 409
    description = 'Conflict'


class ServerError(Error):
    http_status = 500
    description = 'Internal Server Error'


class ValidationError(BadRequest):
    def __init__(self, message, path=None):
        super().__init__(message)
        self.path = path or []

    def format(self) -> ErrorResponse:
        msg = super().format()
        msg.description = self.description
        if self.path:
           msg.path = '.'.join(self.path)
        return msg



from abc import ABCMeta, abstractmethod
from typing import Any, Union, Generic, TypeVar
from typing import Type

from venom.message import Message

T = TypeVar('T', bool, int, float, str, bytes, 'venom.message.Message')

P = TypeVar('P')


class Converter(Generic[T, P], metaclass=ABCMeta):
    wire: Type[T]
    python: Type[P]

    @abstractmethod
    def convert(self, value: T) -> P:
        pass

    @abstractmethod
    def format(self, value: P) -> T:
        pass

    def __repr__(self):
        return '<converter {}[{}, {}]>'.format(self.__class__.__name__, self.python.__name__, self.wire.__name__)


import asyncio
import enum
import re
import warnings
from abc import abstractmethod, ABCMeta, ABC
from functools import partial
from types import MethodType
from typing import Callable, Any, Type, Union, Set, Dict, Sequence, Tuple, Mapping, Optional, Awaitable, NamedTuple, \
    TypeVar, Generic, overload, Coroutine

from venom.converter import Converter
from venom.exceptions import NotImplemented_
from venom.message import Message, Empty, field_names
from venom.rpc.inspection import magic_normalize
from venom.rpc.resolver import Resolver
from venom.util import AttributeDict, MetaDict, cached_property

_RULE_PARAMETER_RE = re.compile('\{([^}:]+)(:[^}]*)?\}')


class HTTPFieldLocation(enum.Enum):
    HEADER = 'header'
    QUERY = 'query'
    PATH = 'path'
    BODY = 'body'


class HTTPVerb(enum.Enum):
    GET = 'GET'
    PUT = 'PUT'
    POST = 'POST'
    PATCH = 'PATCH'
    DELETE = 'DELETE'


Req = TypeVar('Req', bound=Message)
Res = TypeVar('Res', bound=Message)

Service = 'venom.rpc.service.Service'


class MethodDescriptor(Generic[Req, Res], metaclass=ABCMeta):

    def __init__(self,
                 request: Type[Req] = None,
                 response: Type[Res] = None,
                 name: str = None,
                 *,
                 http_path: str = None,
                 http_method: HTTPVerb = None,
                 http_status: int = None,
                 **options: Dict[str, Any]) -> None:
        self.request = request
        self.response = response
        self.name = name

        self.http_path = http_path
        self.http_method = http_method
        self.http_status = http_status
        self.options = options

        self._attr_name = None

    def __set_name__(self, owner: Any, name: str):
        self._attr_name = name

    @overload
    def __get__(self, instance: None, owner: Type[Service] = None) -> 'MethodDescriptor[Req, Res]': pass

    @overload
    def __get__(self, instance: Service, owner: Type[Service] = None) -> 'Callable[[Any, Req], Awaitable[Res]]': pass

    def __get__(self, instance, owner):
        if self._attr_name:
            from .stub import Stub
            from .service import Service
            if issubclass(owner, Stub):
                method = self.stub(owner, self._attr_name)
                setattr(owner, self._attr_name, method)
            elif isinstance(instance, Service):
                method = self.prepare(instance, self._attr_name)
                setattr(instance, self._attr_name, method)
            else:
                return self
            if instance:
                return method.__get__(instance, owner)
            return method
        return self

    def _get_http_path(self, service: Type[Service], name: str):
        if self.http_path is None:
            http_path = f"./{name.lower().replace('_', '-')}"
        elif self.http_path == '':
            http_path = '.'
        else:
            http_path = self.http_path

        if service is not None and http_path.startswith('.'):
            return service.__meta__.http_path + http_path[1:]
        return http_path

    def _get_http_method(self) -> HTTPVerb:
        if self.http_method is None:
            return HTTPVerb.POST
        return self.http_method

    def _get_http_status(self, response: Type[Message]):
        if self.http_status is None:
            if response == Empty:
                return 204  # No Content
            return 200  # OK
        return self.http_status

    def prepare(self, service: Type[Service], name: str) -> 'Method':
        return Method(self.name or name,
                      self.request or Empty,
                      self.response or Empty,
                      service,
                      http_path=self._get_http_path(service, self.name or name),
                      http_method=self._get_http_method(),
                      http_status=self._get_http_status(self.response or Empty),
                      **self.options)

    def stub(self, stub: Type[Service], name: str) -> 'Method':
        return Method(self.name or name,
                      self.request or Empty,
                      self.response or Empty,
                      stub,
                      http_path=self._get_http_path(stub, self.name or name),
                      http_method=self._get_http_method(),
                      http_status=self._get_http_status(self.response or Empty),
                      **self.options)


S = TypeVar('S', bound='venom.rpc.service.Service')


class Method(Generic[S, Req, Res], MethodDescriptor[Req, Res]):
    service: Type[S]

    def __init__(self,
                 name: str,
                 request: Type[Req],
                 response: Type[Res],
                 service: Type[S],
                 *,
                 http_path: str,
                 http_method: HTTPVerb,
                 http_status: int,
                 **options: Dict[str, Any]) -> None:
        super().__init__(request,
                         response,
                         name,
                         http_path=http_path,
                         http_method=http_method,
                         http_status=http_status,
                         **options)
        self.service = service

    def http_path_parameters(self) -> Set[str]:
        return set(m.group(1) for m in re.finditer(_RULE_PARAMETER_RE, self.http_path or ''))

    def http_field_locations(self) -> Dict[HTTPFieldLocation, Set[str]]:
        locations = {location: set() for location in HTTPFieldLocation}

        path = set()
        for name in self.http_path_parameters():
            if name in field_names(self.request):
                path.add(name)

        if path:
            locations[HTTPFieldLocation.PATH] = path

        remaining = set()
        for name in field_names(self.request):
            if name not in path:
                remaining.add(name)

        if self.http_method in (HTTPVerb.POST, HTTPVerb.PATCH, HTTPVerb.PUT):
            locations[HTTPFieldLocation.BODY] = remaining
        else:
            locations[HTTPFieldLocation.QUERY] = remaining
        return locations

    # TODO Error handling. Only errors that are venom.exceptions.Error instances should be raised
    async def invoke(self, instance: S, request: Req, loop: 'asyncio.AbstractEventLoop' = None) -> Res:
        raise NotImplemented_()

    @overload
    def __get__(self, instance: None, owner: Type[S] = None) -> 'MethodDescriptor[Req, Res]': pass

    @overload
    def __get__(self, instance: S, owner: Type[S] = None) -> 'Callable[[S, Req], Awaitable[Res]]': pass

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        else:
            return MethodType(self.invoke, instance)

    def __repr__(self):
        return f'<{self.__class__.__name__} [{self.service.__meta__.name}.{self.name}]>'
        # return f'<{self.__class__.__name__} ' \
        #        f'[{self.name}({self.request.__meta__.name}) -> {self.response.__meta__.name}]>'


class ServiceMethodDescriptor(MethodDescriptor[Req, Res]):
    def __init__(self,
                 func: Callable[..., Any],
                 request: Type[Req] = None,
                 response: Type[Res] = None,
                 name: str = None,
                 *,
                 http_path: str = None,
                 http_method: HTTPVerb = None,
                 http_status: int = None,
                 **options: Dict[str, Any]) -> None:
        super().__init__(request,
                         response,
                         name,
                         http_path=http_path,
                         http_method=http_method,
                         http_status=http_status,
                         **options)
        self._func = func

    def prepare(self,
                service: Type[Service],
                name: str,
                *args: Tuple[Resolver, ...],
                converters: Sequence[Converter] = ()) -> 'ServiceMethod':
        request = self.request
        response = self.response

        if service.__stub__:
            try:
                stub_method = service.__stub__.__methods__[name]
                request = request or stub_method.request
                response = response or stub_method.response

            except KeyError:
                pass  # method not specified in stub

        magic_func = magic_normalize(self._func,
                                     func_name=name,
                                     request=request,
                                     response=response,
                                     owner=service,
                                     additional_args=args,
                                     converters=tuple(converters) + tuple(service.__meta__.converters),
                                     auto_generate_request=self.options.get('auto', False))

        return ServiceMethod(name,
                             magic_func.request,
                             magic_func.response,
                             service,
                             magic_func.invokable,
                             http_path=self._get_http_path(service, name),
                             http_method=self._get_http_method(),
                             http_status=self._get_http_status(magic_func.response),
                             **self.options)

    def stub(self, stub: Type[Service], name: str) -> 'Method':
        magic_func = magic_normalize(self._func,
                                     func_name=name,
                                     request=self.request,
                                     response=self.response,
                                     converters=tuple(stub.__meta__.converters),
                                     auto_generate_request=self.options.get('auto', False))

        return Method(name,
                      magic_func.request,
                      magic_func.response,
                      stub,
                      http_path=self._get_http_path(stub, name),
                      http_method=self._get_http_method(),
                      http_status=self._get_http_status(magic_func.response),
                      **self.options)

    # def __call__(self, *args, **kwargs):
    #     return self._func(*args, **kwargs)


class ServiceMethod(Method[S, Req, Res]):
    implementation: Callable[[S, Req], Awaitable[Res]]

    def __init__(self,
                 name: str,
                 request: Type[Req],
                 response: Type[Res],
                 service: Type[S],
                 implementation: Callable[[S, Req], Awaitable[Res]],
                 *,
                 http_path: str = None,
                 http_method: HTTPVerb = None,
                 http_status: int = None,
                 **options: Dict[str, Any]) -> None:
        super().__init__(name,
                         request,
                         response,
                         service,
                         http_path=http_path,
                         http_method=http_method,
                         http_status=http_status,
                         **options)
        self.implementation = implementation

    def prepare(self, service: Type[Service], name: str) -> 'ServiceMethod':
        return ServiceMethod(name,
                             self.request,
                             self.response,
                             service,
                             self.implementation,
                             http_path=self._get_http_path(service, name),
                             http_method=self._get_http_method(),
                             http_status=self._get_http_status(self.response),
                             **self.options)

    async def invoke(self, instance: S, request: Message, loop: 'asyncio.AbstractEventLoop' = None):
        try:
            return await self.implementation(instance, request, loop=loop)
        except NotImplementedError:
            raise NotImplemented_()


class MethodDecorator(object):
    def __init__(self, descriptor=ServiceMethodDescriptor, **method_options):
        self.descriptor = descriptor
        self.descriptor_kwargs = method_options

    def __call__(self, *args, **kwargs) -> Union[Method, Callable[[Callable], Method]]:
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return self.descriptor(args[0])
        else:
            return lambda fn: self.descriptor(fn, *args, **kwargs)


class HTTPMethodDecorator(MethodDecorator):
    def __call__(self, method: HTTPVerb, *args, **kwargs) -> Union[Method, Callable[[Callable], Method]]:
        if isinstance(method, str):
            verb = HTTPVerb[method]

        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return self.descriptor(args[0], http_method=method, **self.descriptor_kwargs)
        else:
            def _http(path=None, *args, **kwargs):
                return MethodDecorator.__call__(self,
                                                *args,
                                                http_method=method,
                                                http_path=path,
                                                **kwargs,
                                                **self.descriptor_kwargs)

            return _http(*args, **kwargs)

    def GET(self, *args, **kwargs) -> Union[Method, Callable[[Callable], Method]]:
        return self(HTTPVerb.GET, *args, **kwargs)

    def PUT(self, *args, **kwargs) -> Union[Method, Callable[[Callable], Method]]:
        return self(HTTPVerb.PUT, *args, **kwargs)

    def POST(self, *args, **kwargs) -> Union[Method, Callable[[Callable], Method]]:
        return self(HTTPVerb.POST, *args, **kwargs)

    def PATCH(self, *args, **kwargs) -> Union[Method, Callable[[Callable], Method]]:
        return self(HTTPVerb.PATCH, *args, **kwargs)

    def DELETE(self, *args, **kwargs) -> Union[Method, Callable[[Callable], Method]]:
        return self(HTTPVerb.DELETE, *args, **kwargs)


rpc = MethodDecorator()
http = HTTPMethodDecorator()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import codecs
from setuptools import setup, find_packages

setup(
    name='venom',
    version='1.0.0a5',
    packages=find_packages(exclude=['*tests*']),
    url='https://github.com/biosustain/venom',
    license='MIT',
    author='Lars Schöning',
    author_email='lays@biosustain.dtu.dk',
    description='Venom is an upcoming RPC framework for Python',
    long_description=codecs.open('README.rst', encoding='utf-8').read(),
    test_suite='nose.collector',
    tests_require=[
        'aiohttp>=2.0.1',
        'ujson',
        'nose'
    ],
    install_requires=[
        'blinker>=1.4.0'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    zip_safe=False,
    extras_require={
        'docs': ['sphinx'],
        'aiohttp': ['aiohttp>=1.2.0', 'ujson'],
        'grpc': ['grpcio'],
    }
)



from typing import Type, Union, Iterable, Tuple, ClassVar, TypeVar, overload, Mapping
from weakref import WeakKeyDictionary

import asyncio
from blinker import Signal

from venom.rpc.context import RequestContext, DictRequestContext
from venom.rpc.stub import Stub
from .method import rpc, http, Method
from .proxy import ServiceProxy, proxy
from .service import Service


class UnknownService(RuntimeError):
    pass

S = TypeVar('S', bound=Service)


class Venom(object):
    on_add_service: ClassVar[Signal] = Signal('add-service')
    on_add_public_service: ClassVar[Signal] = Signal('add-public-service')
    before_invoke: ClassVar[Signal] = Signal('before-invoke')

    _request_context_cls: Type[RequestContext]
    _instances: Mapping[Service, Union[Service, 'venom.rpc.comms.AbstractClient']]

    def __init__(self, *, request_context_cls: Type[RequestContext] = DictRequestContext, **options):
        self._request_context_cls = request_context_cls
        self._instances = {}
        self._services = {}
        self._public_services = {}
        self._clients = {}
        self.options = options

    # TODO change signature so that all keyword arguments go to the client_cls on init.
    # TODO add internal: bool = None flag; internal = False makes Stubs publicly available.
    def add(self,
            service: Type[Service],
            client: Type['venom.rpc.comms.HTTPClient'] = None,
            *client_args,
            public: bool = None,
            **client_kwargs) -> None:

        name = service.__meta__.name
        if name in self._services:
            if self._services[name] is service:
                return
            raise ValueError(f"A service with name '{name}' already exists")

        self._services[name] = service

        if client:
            self._clients[service] = client(service(self), *client_args, **client_kwargs)
        elif public is None:
            public = True

        service.__manager__.register(self)

        if public:
            self._public_services[name] = service
            self._instances[service] = instance = service(self)
            self.on_add_public_service.send(self, service=instance)

    def _resolve_service_cls(self, reference: Union[str, Type[Service]]):
        if isinstance(reference, str):
            try:
                return self._services[reference]
            except KeyError:
                raise UnknownService(f"No service with name '{reference}' is known to this Venom")
        elif reference not in self._services.values():
            raise UnknownService("'{}' is not known to this Venom".format(reference))
        return reference

    @overload
    def get_instance(self, reference: Type[S]) -> S: pass

    @overload
    def get_instance(self, reference: str) -> Service: pass

    def get_instance(self, reference):
        cls = self._resolve_service_cls(reference)

        try:
            return self._instances[cls]
        except KeyError:
            pass

        if issubclass(cls, Stub):
            instance = self._clients[cls]
        else:
            raise RuntimeError
            # self._instances[cls] = instance = cls(venom=self)

        return instance

    def iter_methods(self) -> Iterable[Method]:
        for instance in self._instances.values():
            for method in instance.__methods__.values():
                yield method

    def get_request_context(self) -> RequestContext:
        return self._request_context_cls(self)

    async def _invoke(self, method: Method, request: 'venom.Message'):
        with self._request_context_cls(self):
            instance = self.get_instance(method.service)
            self.before_invoke.send(self, method=method, request=request)
            return await method.invoke(instance, request)

    async def invoke(self, method: Method, request: 'venom.Message', loop: 'asyncio.AbstractEventLoop' = None):
        if loop is None:
            loop = asyncio.get_event_loop()
        return await loop.create_task(self._invoke(method, request))

    def __iter__(self) -> Iterable[Type[Service]]:
        return iter(self._instances.values())




from typing import Any

import datetime

from venom.common import DateConverter, DateTimeConverter
from venom.fields import ConverterField


class Date(ConverterField):
    def __init__(self, **kwargs) -> None:
        super().__init__(DateConverter(), **kwargs)


class DateTime(ConverterField):
    def __init__(self, **kwargs) -> None:
        super().__init__(DateTimeConverter(), **kwargs)



from typing import Union, Type, TypeVar, Generic, cast, Any
from venom.rpc.service import Service

S = TypeVar('S', bound=Service)


class ServiceProxy(Generic[S]):
    # TODO define life-time. can be global or request or session.
    def __init__(self, reference: Union[str, Type[S]]) -> None:
        self.reference = reference

    def __get__(self, service: Service, owner: Any = None) -> S:
        return cast(S, service.venom.get_instance(self.reference))


def proxy(service: Type[S]) -> S:
    return cast(S, ServiceProxy(service))



from abc import abstractmethod, ABCMeta
from typing import TypeVar, Generic, Type

T = TypeVar('T')


class Resolver(Generic[T], metaclass=ABCMeta):
    python: Type[T] = None

    async def __call__(self,
                       service: 'venom.rpc.service.Service',
                       request: 'venom.message.Message') -> T:
        return await self.resolve(service, request)

    @abstractmethod
    async def resolve(self,
                      service: 'venom.rpc.service.Service',
                      request: 'venom.message.Message') -> T:
        pass



from typing import Dict, Any, Tuple


# FIXME should be Generic
class AttributeDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


def _meta_obj_to_dict(meta_obj):
    dct = {}
    for k, v in meta_obj.__dict__.items():
        if not k.startswith('__'):
            dct[k] = v
    return dct

# FIXME should be AttributeDict[str, Any]
MetaDict = AttributeDict


def meta(bases, members, meta_name='Meta') -> Tuple[MetaDict, MetaDict]:
    meta_ = AttributeDict()
    for base in bases:
        if hasattr(base, '__meta__') and base.__meta__ is not None:
            meta_.update(base.__meta__)
        elif hasattr(base, meta_name):
            meta_.update(_meta_obj_to_dict(getattr(base, meta_name)))

    changes = {}
    if meta_name in members:
        changes = _meta_obj_to_dict(members[meta_name])
        meta_.update(changes)
    return meta_, AttributeDict(changes)


def upper_camelcase(s: str) -> str:
    return s.title().replace('_', '')


def camelcase(s):
    return s[0].lower() + s.title().replace('_', '')[1:] if s else s


class cached_property(object):
    """
    Descriptor (non-data) for building an attribute on-demand on first use.
    """
    def __init__(self, factory):
        """
        <factory> is called such: factory(instance) to build the attribute.
        """
        self._attr_name = factory.__name__
        self._factory = factory

    def __get__(self, instance, owner):
        # Build the attribute.
        attr = self._factory(instance)

        # Cache the value; hide ourselves.
        setattr(instance, self._attr_name, attr)

        return attr



from venom import Message
from venom.rpc import Stub
from venom.rpc import http
from venom.fields import Field, Repeat, String, Map, Bool


class SchemaMessage(Message):
    type = String()  # TODO: enum
    description = String()
    properties = Map(Field('venom.rpc.reflect.openapi.SchemaMessage'))
    ref = String(json_name='$ref')
    additional_properties = Field('venom.rpc.reflect.openapi.SchemaMessage')
    items_ = Field('venom.rpc.reflect.openapi.SchemaMessage', name='items')


class ParameterMessage(Message):
    is_in = String(json_name='in')  # TODO: enum
    description = String()
    required = Bool()
    name = String()
    type = String()
    items_ = Field(SchemaMessage, name='items')
    schema = Field(SchemaMessage)


class ResponseMessage(Message):
    description = String()
    schema = Field(SchemaMessage)


class ResponsesMessage(Message):
    default = Field(ResponseMessage)  # TODO: error codes


class OperationMessage(Message):
    produces = Repeat(String())
    responses = Field(ResponsesMessage)
    parameters = Repeat(ParameterMessage)


class InfoMessage(Message):
    version = String()
    title = String()
    description = String()
    terms_of_service = String()
    contact = String()
    license = String()


class OpenAPISchema(Message):
    swagger = String()
    schemes = Repeat(String())
    consumes = Repeat(String())
    produces = Repeat(String())
    info = Field(InfoMessage)
    host = String()
    base_path = String()
    paths = Map(Map(Field(OperationMessage)))
    definitions = Map(Field(SchemaMessage))


class ReflectStub(Stub):

    @http.GET('/openapi.json')
    def get_openapi_schema(self) -> OpenAPISchema:
        raise NotImplementedError()
